library(caret)

# ----1.----

data = read.csv("parkinsons.csv")

# Split data
n = dim(data)[1]
set.seed(12345)
id = sample(1:n, floor(n * 0.6))
train = data[id, ]
test = data[-id, ]

# Scale data
scaler <- preProcess(train)
train_scaled <- predict(scaler, train)
test_scaled <- predict(scaler, test)

y_train <- train_scaled[, 5]
X_train <- train_scaled[, -(1:6)]
y_test <- test_scaled[, 5]
X_test <- test_scaled[, -(1:6)]

# ----2.----

# Mean square error
mse <- function(y, y_hat)
  mean((y_hat - y) ^ 2)

feature_cols <- colnames(X_train)
target_col <- colnames(train_scaled[5])

formula <- as.formula(paste(target_col, "~", paste(feature_cols, collapse = " + "), "-1"))
model <- lm(formula, data = train_scaled)

model_summary <- summary(model)

print(model_summary)

theta <- model_summary$coefficients[, 1]
sigma = model_summary$sigma

train_pred <- predict(model, train_scaled)
test_pred <- predict(model, test_scaled)

print(paste(
  "MSE on the training data:",
  mse(train_scaled$motor_UPDRS, train_pred)
))
print(paste("MSE on the test data:", mse(test_scaled$motor_UPDRS, test_pred)))

# ----3.----

# Loglikelihood
log_likelihood <- function(theta, sigma) {
  n <- length(y_train)
  ll <- -n / 2 * log(2 * pi) - n * log(sigma) - (1 / (2 * sigma ^ 2)) * sum((y_train -
                                                                               as.matrix(X_train) %*% theta) ^
                                                                              2)
  return(ll)
}

# Ridge
ridge <- function(theta, sigma, lambda) {
  ridge <- -log_likelihood(theta, sigma) + lambda * sum(theta ^ 2)
  return(ridge)
}

# RidgeOpt
ridge_opt <- function(lambda) {
  objective_function <- function(params) {
    theta <- params[1:(length(params) - 1)]
    log_sigma <- params[length(params)] # Using log to avoid negative sigma
    sigma <- exp(log_sigma)
    return(ridge(theta, sigma, lambda))
  }
  
  return(optim(
    par = rep(0, 17),
    fn = objective_function,
    method = "BFGS"
  ))
  
}

# DF
df = function(lambda) {
  X <- as.matrix(X_train)
  P <- X %*% solve(t(X) %*% X + lambda * diag(ncol(X))) %*% t(X)
  return(sum(diag(P)))
}


# ----4.----

for (lambda in c(1, 100, 1000)) {
  theta_opti <- unlist(ridge_opt(lambda)$par[-17])
  y_hat_train <- as.matrix(X_train) %*% theta_opti
  y_hat_test <- as.matrix(X_test) %*% theta_opti
  print(paste(
    "MSE for lambda =",
    lambda,
    "on training data:",
    mse(y_train, y_hat_train)
  ))
  print(paste(
    "MSE for lambda =",
    lambda,
    "on test data:",
    mse(y_test, y_hat_test)
  ))
}


for (lambda in c(1, 100, 1000)) {
  print(paste("DF for lambda =", lambda, ":", df(lambda)))
}
